@@ -1,6 +1,9 @@
 from copy import deepcopy as dcopy
+from functools import partial
+import inspect
 import logging
 import multiprocessing
+import warnings
 
 import numpy as np
 import sklearn.metrics as metrics
@@ -58,6 +61,7 @@ def reproducible_worker_dict():
     g.manual_seed(0)
     return {'worker_init_fn': seed_worker, 'generator': g}
 
+# DEPRECATED: use get_forward_f instead
 def _test_assertion(dataset, targets_type, forward_f, result_dict, keep_x):
     # Safety check
     if forward_f is not None and result_dict is not None:
@@ -69,13 +73,13 @@ def _test_assertion(dataset, targets_type, forward_f, result_dict, keep_x):
                 targets_type = dataset.targets_type
             else:
                 raise Exception('When forward_f and result_dict is not given, targets_type must be given or the dataset must have the attribute "targets_type"')
-        assert targets_type in targets_type_list, f'targets_type must be one of {targets_type_list}, received: {targets_type}'
         forward_f, result_dict = get_forward_f(targets_type, keep_x=keep_x)
     else:
         raise Exception(f'forward_f and result_dict must either both be provided or None, received [forward_f: {forward_f}][result_dict: {result_dict}]')
     return forward_f, result_dict
 
-def forward_network(network, dataset, forward_f=None, batch_size=64, targets_type=None, result_dict=None, keep_x=False, num_workers=0, amp=False):
+# def forward_network(network, dataset, forward_f=None, batch_size=64, targets_type=None, result_dict=None, keep_x=False, num_workers=0, amp=False):
+def forward_network(network, dataset, forward_f=None, batch_size=64, targets_type=None, num_workers=0, amp=False):
     """
     Forward pass of the network on the dataset.
     Takes care of device, dataloader, and mixed precision.
@@ -87,17 +91,27 @@ def forward_network(network, dataset, forward_f=None, batch_size=64, targets_typ
     dataset : torch.utils.data.Dataset
         Dataset to be tested.
     forward_f : callable, default=None
-        Function to be tested.
-        If None, it will be inferred from dataset.targets_type
-        (Arguments: data, network, device)
-        (Returns: dictionary of tensors or tuple of tensors)
+        forward pass function to pass the data through the network.
+        If None, either inferred from dataset.targets_type or the data is just fed to the network and returned as output of the network (probably tensor/tuple not dict)
+
+        (Arguments: network, data[, device])
+        (Returns: dictionary or tuple of tensors)
 
     Returns
     -------
     result : ndarrays corresponding to the outputs from the forward_f 
         returns tuple or dict, based on forward_f return
     """
-    forward_f, result_dict = _test_assertion(dataset=dataset, targets_type=targets_type, forward_f=forward_f, result_dict=result_dict, keep_x=keep_x)
+
+    # forward_f check
+    if forward_f is None:
+        if targets_type is None and hasattr(dataset, 'targets_type'):
+            targets_type = dataset.targets_type
+            forward_f = get_forward_f(targets_type)
+            
+        else:
+            forward_f = _forward
+    forward_f = partial(forward_f, network=network)
 
     device = T.torch.get_device(network) # Test on the network's device
     network.eval()
@@ -112,133 +126,195 @@ def forward_network(network, dataset, forward_f=None, batch_size=64, targets_typ
         if amp:
             with torch.autocast(device_type=device.type):
                 for data in test_loader:
-                    result = forward_f(data=data, network=network, device=device, keep_x=keep_x)
+                    data = T.torch.to(data, device)
+                    result = function_call_type_matching(f=forward_f, args=data)
+                    result = T.torch.to(result, device='cpu')
                     l_results.append(result)
         else:
             for data in test_loader:
-                result = forward_f(data=data, network=network, device=device, keep_x=keep_x)
+                data = T.torch.to(data, device)
+                result = function_call_type_matching(f=forward_f, args=data)
+                result = T.torch.to(result, device='cpu')
                 l_results.append(result)
 
     # Formatting: Concatenate, (optional amp), numpy
-    return_type = type(result)
-    if return_type is dict:
+    if isinstance(result, dict): 
         result = T.merge_dict(l_results)
         result = {k: torch.cat(v, dim=0) if len(v)>0 else v for k, v in result.items()}
         if amp: result = {k: v.to(torch.float16) for k, v in result}
         result = {k: v.numpy() for k, v in result.items()}
-    elif return_type is tuple:
+    elif isinstance(result, (list, tuple)):
         result = T.merge_tuple(l_results)
         result = tuple(torch.cat(v, dim=0) if len(v)>0 else v for v in result)
         if amp: result = tuple(v.to(torch.float16) for v in result)
         result = tuple(v.numpy() for v in result)
-    elif return_type is torch.Tensor: # Single tensor return
+    elif isinstance(result, torch.Tensor): # Single tensor return
         result = torch.cat(l_results, dim=0)
         if amp: result = result.to(torch.float16)
         result = result.numpy()
     else:
-        raise Exception(f'Unknown return type: {return_type}')
+        raise Exception(f'Unknown data type from forward_f return: {type(result)}')
 
     return result
 
-def forward_model(model, dataset, batch_size=64, targets_type=None, forward_f=None, result_dict=None, keep_x=False, num_workers=0, amp=False):
+def forward_model(model, dataset, forward_f=None, batch_size=64, targets_type=None, num_workers=0, amp=False):
     '''
     wrapper around model.
     transfers get_f (preprocessing modules) of model to dataset, and returns them to original dataset after inference.
     '''
     network = model.network
     dataset, misc_temp = transfer_misc(model, dataset)
-    result_dict = forward_network(network, dataset, batch_size, targets_type, forward_f, result_dict, keep_x, num_workers, amp)
+    result = forward_network(network=network, dataset=dataset, forward_f=forward_f, batch_size=batch_size, targets_type=targets_type, num_workers=num_workers, amp=amp)
     dataset = inverse_transfer_misc(misc_temp, dataset)
 
-    return result_dict
+    return result
 
-def evaluate(results, metrics):
-    if type(metrics) is dict:
-        scores = {}
-        for metric_name, metric in metrics.items():
-            scores[metric_name] = metric(results)
-    elif type(metrics) is list or type(metrics) is tuple:
-        scores = [metric(results) for metric in metrics]
-    else:
-        scores = metrics(results)
+# def multicall(result, functions):
+def evaluate(result, metrics):
+    if isinstance(metrics, dict):
+        try:
+            scores = {}
+            for metric_name, metric in metrics.items():
+                scores[metric_name] = function_call_type_matching(f=metric, args=result)
+                # if isinstance(result, dict):
+                #     # Only pass the arguments that are in the metric function
+                #     kwargs = {k:v for k, v in result.items() if k in inspect.signature(metric).parameters.keys()}
+                #     scores[metric_name] = metric(**kwargs)
+                # elif isinstance(result, (list, tuple)):
+                #     scores[metric_name] = metric(*result)
+                # else:
+                #     scores[metric_name] = metric(result)
+        except Exception as e:
+            if isinstance(result, dict):
+                message = f'Error at metric: {metric}, result.keys(): {result.keys()}\n'
+            elif isinstance(result, (list, tuple)):
+                message = f'Error at metric: {metric}, len(result): {len(result)}\n'
+            else:
+                message = f'Error at metric: {metric}, type(result): {type(result)}\n'
+            e.add_note(message)
+            raise e
+    elif isinstance(metrics, (list, tuple)):
+        scores = []
+        for metric in metrics:
+            scores.append(function_call_type_matching(f=metric, args=result))
+            # if isinstance(result, dict):
+            #     kwargs = {k:v for k, v in result.items() if k in inspect.signature(metric).parameters.keys()}
+            #     scores.append(metric(**kwargs))
+            # elif isinstance(result, (list, tuple)):
+            #     scores.append(metric(*result))
+            # else:
+            #     scores.append(metric(result))
+    else: # Single metric
+        scores = function_call_type_matching(f=metrics, args=result)
+        # if isinstance(result, dict):
+        #     kwargs = {k:v for k, v in result.items() if k in inspect.signature(metrics).parameters.keys()}
+        #     if len(kwargs) < len(result):
+        #         warnings.warn('Some arguments in result are not used in the metric function, when only single metric was passed')
+        #     scores = metrics(**kwargs)
+        # elif isinstance(result, (list, tuple)):
+        #     scores = metrics(*result)
+        # else:
+        #     scores = metrics(result)
     return scores
 
 targets_type_list = [None, 'categorical', 'multihead_classification', 'autoencode', 'regression'] # move this to utils?
-def get_forward_f(targets_type, keep_x=False):
+def get_forward_f(targets_type):
     assert targets_type in targets_type_list, f'targets_type must be one of {targets_type_list}, received: {targets_type}'
     if targets_type is None or targets_type == 'regression':
         forward_f = _forward_base
-        result_list = ['y_true', 'y_hat']
+        result_keys = ['y_true', 'y_pred']
     elif targets_type == 'categorical':
         forward_f = _forward_categorical
-        result_list = ['y_true', 'y_hat', 'y_score', 'y_pred']
+        result_keys = ['y_true', 'y_hat', 'y_score', 'y_pred'] # y_hat vs y_logit?
     elif targets_type == 'multihead_classification':
         forward_f = _forward_multihead_categorical
-        result_list = ['y_true', 'y_hat', 'y_score', 'y_pred']
+        result_keys = ['y_true', 'y_hat', 'y_score', 'y_pred']
     elif targets_type == 'autoencode':
         forward_f = _forward_autoencode
-        result_list = ['x', 'z', 'x_hat']
+        result_keys = ['x', 'z', 'x_hat']
     else:
         raise Exception(f'unknown targets_type: {targets_type}')
     
-    if keep_x and 'x' not in result_list: result_list.append('x')
-    result_dict = {r:[] for r in result_list}
+    log.info(f'get_forward_f: targets_type: {targets_type}, keys in result (dict): {result_keys}')
 
-    return forward_f, result_dict
+    return forward_f
 
-def _forward_base(data, network, result_dict, device=None, keep_x=False):
+def function_call_type_matching(f, args):
+    if isinstance(args, dict):
+        function_arguments = inspect.signature(f).parameters.keys()
+        if 'kwargs' not in function_arguments: # Will raise error if undefined arguments are passed
+            # Only pass the arguments that are in the function
+            kwargs = {k:v for k, v in args.items() if k in function_arguments}
+        else: # kwargs will allow undefined arguments of f
+            kwargs = args
+        return f(**kwargs)
+    elif isinstance(args, (list, tuple)):
+        return f(*args)
+    else:
+        return f(args)
+
+def _forward(network, data):
+    return network(data)
+    # if isinstance(data, dict):
+    #     return network(**data)
+    # elif isinstance(data, (list, tuple)):
+    #     return network(*data)
+    # else:
+    #     return network(data)
+
+def _forward_base(network, data):
     # device = device if device is not None else T.torch.get_device(network)
-    x = data['x'].to(device)
-    y = data['y'].to(device)
-    y_hat = network(x)
+    x = data['x']
+    y = data['y']
+    y_pred = network(x)
 
     result_dict = {
-        'y_true': y.cpu(),
-        'y_hat': y_hat.cpu()
+        'y_true': y,
+        'y_pred': y_pred
     }
-    if keep_x: result_dict['x'] = x.cpu()
+    # if keep_x: result_dict['x'] = x
     return result_dict
 
-def _forward_categorical(data, network, result_dict, device=None, keep_x=False):
+def _forward_categorical(network, data):
     # Is y_hat necessary? for torch_crossentropyloss? any other methods?
-    x = data['x'].to(device)
-    y = data['y'].to(device)
+    x = data['x']
+    y = data['y']
     y_hat = network(x)
     y_score = torch.softmax(y_hat, dim=1) # (N, n_classes)
 
     result_dict = {
-        'y_true': y.cpu(),
-        'y_hat': y_hat.cpu(),
-        'y_score': y_score.cpu(),
-        'y_pred': y_score.argmax(axis=1).cpu()
+        'y_true': y,
+        'y_hat': y_hat,
+        'y_score': y_score,
+        'y_pred': y_score.argmax(axis=1)
     }
-    if keep_x: result_dict['x'] = x.cpu()
+    # if keep_x: result_dict['x'] = x
     return result_dict
 
-def _forward_multihead_categorical(data, network, result_dict, device=None, keep_x=False):
-    x = data['x'].to(device)
-    y = data['y'].to(device)
+def _forward_multihead_categorical(network, data):
+    x = data['x']
+    y = data['y']
     y_hat = network(x)
     y_score = torch.softmax(y_hat, dim=-1) # (N, subtype, n_classes)
 
     result_dict = {
-        'y_true': y.cpu(),
-        'y_hat': y_hat.cpu(),
-        'y_score': y_score.cpu(),
-        'y_pred': y_score.argmax(axis=-1).cpu()
+        'y_true': y,
+        'y_hat': y_hat,
+        'y_score': y_score,
+        'y_pred': y_score.argmax(axis=-1)
     }
-    if keep_x: result_dict['x'] = x.cpu()
+    # if keep_x: result_dict['x'] = x
     return result_dict
 
-def _forward_autoencode(data, network, result_dict, device=None):
-    x = data['x'].to(device)
+def _forward_autoencode(network, data):
+    x = data['x']
     z = network.encoder(x)
     x_hat = torch.sigmoid(network.decoder(z))
 
     result_dict = {
-        'x': x.cpu(),
-        'z': z.cpu(),
-        'x_hat': x_hat.cpu()
+        'x': x,
+        'z': z,
+        'x_hat': x_hat
     }
     return result_dict
 
