@@ -29,41 +29,77 @@ from . import model as M
 log = logging.getLogger(__name__)
 
 # %%
+def validate_cfg_val(cfg_val):
+    cfg_val = dcopy(dict(cfg_val)) if cfg_val is not None else {}
+    if 'forward_f' not in cfg_val:
+        cfg_val['forward_f'] = None
+    return cfg_val
+
 class Trainer:
-    def __init__(self, network, train_dataset=None, cfg_train={}, criterion=None, network_dir=None, cfg_val=None, val_dataset=None, val_metrics=None, name='default', verbose=True, amp=False, reproduce=True):
-        '''
+    def __init__(self, network, train_dataset=None, cfg_train={}, criterion=None, network_dir='network', cfg_val=None, val_dataset=None, val_metrics=None, name='default', verbose=True, amp=False, reproduce=True):
+        """
         - train network with early stopping, given hyperparameters
-        - save/load network with necessary preprocessing pipeline
+        - save/load network with necessary preprocessing pipelin
 
-        :param network: torch.nn.Module object
-        :param train_dataset: torch.utils.data.Dataset object
+        Parameters
+        ----------
+        network : torch.nn.Module object
+            Network to train
+
+        train_dataset: torch.utils.data.Dataset object
             This dataset is used for training.
-            Recommended return format from dataset is: {'x': x, 'y': y}
-            where 'x' is the input, and 'y' is the target values.
+            Recommended return format from dataset is dictionary form, although tuples or raw tensors are supported.
+            Example: {'x': x, 'y': y}
+            
+            train_dataset is fed into torch.utils.data.DataLoader object, where it is batched then passed to self.forward(), with the following pattern depending on train_dataset return type:
+                torch.Tensor: self.forward(data)
+                tuple or list: self.forward(*data)
+                dict: self.forward(**data)
 
-        :param validation: dict of Validation objects
+        val_dataset: torch.utils.data.Dataset object or dict of such objects
+            Dataset used for validation (Early stopping)
+            forward mechanism from val_dataset may be different from train_dataset, which could be defined in cfg_val['forward_f']        
 
-        :param cfg_train: dict-like object which contains:
-            lr
-            batch_size
-            weight_decay (optional)
-            cv_step (optional)
+        cfg_train: dict-like object with following keys:
+            lr: float
+            batch_size: int
+            weight_decay: bool (optional)
+            cv_step: int (optional, recommended when val_dataset is given)
         
-        :param cfg_val (optional): dict-like object which contains:
-            increase_better
-            patience
-            batch_size
-            target_dataset (optional, if val_dataset is dict for multiple datasets)
-            criterion (optional, if val_metrics is dict for multiple metrics) 
+        cfg_val: (required if val_dataset is given) dict-like object with following keys:
+            increase_better: Bool
+            patience: int
+            batch_size: int
 
-        :param criterion: torch.nn.Module object, used to compute loss function
+            forward_f: callable, forward pass that might be different from training forward pass (optional)
+            target_dataset: str (optional, if val_dataset is dict for multiple datasets)
+            criterion: str (optional, if val_metrics is dict for multiple metrics) 
+        
+        criterion: nn.Module object
+            Defines the loss function.
+            If parameters exist within criterion, it is moved to the corresponding device where network lies.
+        
+        network_dir: str or Pathlike object, default='network'
+            The directory to save the nn.
+            If early stopping is used, a subdirectory will be created that keeps the models during training.
 
-        Recommended way of making hs.node.Node object is like the following:
+        Examples
+        --------
+        # Recommended way of making hs.Trainer object is like the following:
+
+        kwargs = {'network': Network(), 'train_dataset': train_dataset, 'validation': None, 'cfg_train': cfg.train,
+                'criterion': criterion, 'network_dir': path['network'], 'verbose': True, 'amp': True}
+        trainer = hs.Trainer(**kwargs)
+
+        # To make a custom Trainer for different neural network applications, just define the forward function:
+
+        class MyTrainer(hs.Trainer):
+            def forward(self, x, y, z):
+                w = self.network(x,y)
+                loss = self.criterion(w, z)
+                return loss
+        """
 
-            kwargs = {'network': Network(), 'train_dataset': train_dataset, 'validation': None, 'cfg_train': cfg.train,
-                    'criterion': criterion, 'network_dir': path['network'], 'verbose': True, 'amp': True}
-            node = hs.Trainer(**kwargs)
-        '''
         # Store configurations
         self.network = network
         self.train_dataset = train_dataset
@@ -71,10 +107,10 @@ class Trainer:
         self.criterion = criterion
         self.network_dir = network_dir
 
-        self.cfg_val = dcopy(dict(cfg_val)) if cfg_val is not None else {}
+        self.cfg_val = validate_cfg_val(cfg_val)
         self.val_dataset = val_dataset
         self.val_metrics = val_metrics
-        self.earlystopper = EarlyStopper(increase_better=cfg_val['increase_better'], patience=cfg_val['patience']) if self.val_dataset is not None else None
+        self.earlystopper = EarlyStopper(increase_better=cfg_val['increase_better'], patience=cfg_val['patience'], save_dir=os.path.join(self.network_dir, 'network_temp')) if self.val_dataset is not None else None
         if 'target_dataset' in self.cfg_val:
             if type(self.val_dataset)==dict:
                 assert self.cfg_val['target_dataset'] in self.val_dataset, f'[Validation setup] target_dataset: {self.cfg_val["target_dataset"]} not found in val_dataset: {self.val_dataset.keys()}'
@@ -123,22 +159,24 @@ class Trainer:
             if type(self.val_dataset)==dict:
                 scores = {}
                 for name, dataset in self.val_dataset.items():
-                    d_results = E.forward_model(self.model, dataset, batch_size=self.cfg_val['batch_size'], amp=self.amp)
-                    if any([np.isnan(v).any() for v in d_results.values()]):
-                        log.warning(f'[Node: {self.name}][validate] NaN found in d_results. Skipping validation...')
+                    results = E.forward_model(self.model, dataset, batch_size=self.cfg_val['batch_size'], forward_f=self.cfg_val['forward_f'], amp=self.amp)
+                    isnan = any([np.isnan(v).any() for v in results.values()]) if isinstance(results, dict) else np.isnan(np.concatenate(results)).any()
+                    if isnan:
+                        log.warning(f'[Node: {self.name}][validate] NaN found in results. Skipping validation...')
                         score = {self.cfg_val['criterion']: self.earlystopper.best_score}
                     else:
-                        score = E.evaluate(d_results, self.val_metrics)
+                        score = E.evaluate(results, self.val_metrics)
                     scores[name] = score
                     self.print(f'[dataset_name: {name}] Validation Score: {score}')
                 score = scores[self.cfg_val['target_dataset']]
             else:
-                d_results = E.forward_model(self.model, self.val_dataset, batch_size=self.cfg_val['batch_size'], amp=self.amp)
-                if any([np.isnan(v).any() for v in d_results.values()]):
-                    log.warning(f'[Node: {self.name}][validate] NaN found in d_results. Skipping validation...')
+                results = E.forward_model(self.model, self.val_dataset, batch_size=self.cfg_val['batch_size'], forward_f=self.cfg_val['forward_f'], amp=self.amp)
+                isnan = any([np.isnan(v).any() for v in results.values()]) if isinstance(results, dict) else np.isnan(np.concatenate(results)).any()
+                if isnan:
+                    log.warning(f'[Node: {self.name}][validate] NaN found in results. Skipping validation...')
                     score = {self.cfg_val['criterion']: self.earlystopper.best_score}
                 else:
-                    score = E.evaluate(d_results, self.val_metrics)
+                    score = E.evaluate(results, self.val_metrics)
                 self.print(f'Validation Score: {score}')
             
             if self.earlystopper is not None:
@@ -153,6 +191,7 @@ class Trainer:
                     self.earlystopper.history.reset()
                     
         torch.cuda.empty_cache()
+        self.network.train()
         return patience_end
 
     def generate_loader(self):
@@ -160,8 +199,6 @@ class Trainer:
             reproduce_kwargs = U.reproducible_worker_dict() if self.reproduce else {}
             drop_last = len(self.train_dataset) % self.cfg_train['batch_size'] == 1 # To avoid batch normalization layers from raising exceptions
             self.loader = D.DataLoader(self.train_dataset, batch_size=self.cfg_train['batch_size'], shuffle=True, drop_last=drop_last, **reproduce_kwargs)
-            # self.loader = D.DataLoader(self.train_dataset, batch_size=self.cfg_train['batch_size'], shuffle=True, drop_last=True, **U.reproducible_worker_dict()) if len(self.train_dataset) % self.cfg_train['batch_size'] == 1 \
-            #         else D.DataLoader(self.train_dataset, batch_size=self.cfg_train['batch_size'], shuffle=True, drop_last=False, **U.reproducible_worker_dict())
         else:
             log.warning('No train_dataset found. Skipping generate_loader()')
 
@@ -214,7 +251,6 @@ class Trainer:
             step = step
             log.debug(f'[Node: {self.name}] train for {step} steps')
 
-        self.network.train()
         self._device = device if device is not None else tools.torch.get_device(self.network)
         self.criterion = self.criterion.to(self._device)
         if reset_loss_tracker: self.loss_tracker.reset()
@@ -226,7 +262,7 @@ class Trainer:
 
         if self.val_dataset is not None:
             if 'cv_step' not in self.cfg_train:
-                self.print('Node.validation given, but "cv_step" not specified in cfg_train. Defaults to 1 epoch')
+                self.print('val_dataset is given but "cv_step" not specified in cfg_train. Defaults to 1 epoch')
                 self.cfg_train['cv_step'] = len(self.loader)
             self.validate() # initial testing
 
@@ -240,9 +276,9 @@ class Trainer:
             self.op = optim.Adam(self.network.parameters(), **kwargs)
 
         if horizon == 'epoch':
-            self._update_epoch(T=epoch, no_val=no_val, device=self._device)
+            self._update_epoch(T=epoch, no_val=no_val)
         elif horizon=='step':
-            self._update_step(T=step, no_val=no_val, device=self._device)
+            self._update_step(T=step, no_val=no_val)
 
         # TODO: Return criterion back to its original device, meaning we have to store its previous device info
         self.criterion = self.criterion.cpu()
@@ -318,25 +354,20 @@ class Trainer:
         -------
         loss : float
         """
-        try:
-            # Automatic Mixed Precision (AMP, float32 -> float16) for acceleration
-            if self.amp:
-                with torch.autocast(device_type=self._device.type):
-                    loss, N = self._forward(data)
-            else:
+        # Automatic Mixed Precision (AMP, float32 -> float16) for acceleration
+        if self.amp:
+            with torch.autocast(device_type=self._device.type):
                 loss, N = self._forward(data)
+        else:
+            loss, N = self._forward(data)
 
-            self.op.zero_grad()
-            loss.backward()
-            self.op.step()
+        self.op.zero_grad()
+        loss.backward()
+        self.op.step()
 
-            loss = loss.item()
-            self.train_meter.step(loss, N)
-            return loss
-
-        except Exception as e:
-            log.warning(e)
-            import pdb; pdb.set_trace()
+        loss = loss.item()
+        self.train_meter.step(loss, N)
+        return loss
 
     def _forward(self, data): # TODO: change to function decorator?
         '''
@@ -357,8 +388,13 @@ class Trainer:
             N = len(next(iter(data.values()))) # number of data in batch
             loss = self.forward(**data)
 
+        elif datatype is torch.Tensor:
+            data = data.to(self._device)
+            N = len(data)
+            loss = self.forward(data)
+            
         else:
-            raise Exception(f'return type from dataset must be one of [tuple, list, dict], received: {datatype}')
+            raise Exception(f'return type from dataset must be one of [torch.Tensor, tuple, list, dict], received: {datatype}')
         return loss, N
 
     def forward(self, x, y):
